/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.muk.transfer;

import static de.ozgcloud.nachrichten.postfach.muk.transfer.MukPostfachMessageMapper.*;

import java.io.ByteArrayInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Stream;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.bson.types.ObjectId;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Service;

import de.ozgcloud.apilib.file.OzgCloudFile;
import de.ozgcloud.apilib.file.OzgCloudUploadFile;
import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.muk.elster.transfer.AntwortDetails;
import de.ozgcloud.muk.elster.transfer.AntwortDetailsList;
import de.ozgcloud.muk.elster.transfer.AuftragAngelegt;
import de.ozgcloud.muk.elster.transfer.BereitstellungAnhangWithUploadId;
import de.ozgcloud.muk.elster.transfer.BereitstellungAntwort;
import de.ozgcloud.muk.elster.transfer.BereitstellungAuftragNeuV3;
import de.ozgcloud.muk.elster.transfer.BereitstellungInfo;
import de.ozgcloud.muk.elster.transfer.DownloadAnhangDetails;
import de.ozgcloud.muk.elster.transfer.UploadInfo;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.muk.MukPostfachProperties;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Service
@RequiredArgsConstructor
@ConditionalOnProperty(prefix = MukPostfachProperties.PREFIX, name = { "server" })
public class ElsterTransferRemoteService {
	public static final String POSTFACH = "postfach";
	public static final String VORGANG_ID_REFIX = "VorgangId[";
	public static final String ANTWORTEN_ENDPOINT = "/rest/eingang/antworten";
	public static final String BEREITSTELLUNGS_ENDPUNKT = "/rest/bereitstellungsauftrag/v3";
	public static final String BEREITSTELLUNG_UPLOAD_ENDPUNKT = "/rest/bereitstellungsauftrag/file-upload";
	static final Set<PostfachNachricht.ReplyOption> REPLY_POSSIBLE_OPTION = EnumSet.of(PostfachNachricht.ReplyOption.POSSIBLE,
	  PostfachNachricht.ReplyOption.MANDATORY);
	static final Set<String> MITTEILUNG_DATENARTEN = Set.of(DATENART_KURZMITTEILUNG, DATENART_MITTEILUNG);
	static final Set<String> NACHRICHTEN_MIT_ANHANG = Set.of(DATENART_BESCHEID, DATENART_MITTEILUNG);
	private final ElsterRestClient restClient;
	private final MukPostfachProperties postfachProperties;
	private final MukPostfachMessageMapper mapper;
	private final MukOzgCloudFileMapper fileMapper;
	private final AttachmentService attachmentService;

	private static boolean anhangNeedToBeUploaded(final BereitstellungAuftragNeuV3 bereitstellungAuftrag) {
		return NACHRICHTEN_MIT_ANHANG.contains(bereitstellungAuftrag.getDatenart());
	}

	private static boolean isMitteilung(final BereitstellungAuftragNeuV3 bereitstellungAuftrag) {
		return MITTEILUNG_DATENARTEN.contains(bereitstellungAuftrag.getDatenart());
	}

	private static boolean isReplyPossible(final PostfachNachricht nachricht) {
		return REPLY_POSSIBLE_OPTION.contains(nachricht.getReplyOption());
	}

	static OzgCloudUploadFile createOzgCloudUploadFile(final DownloadAnhangDetails downloadAnhangDetails, final String vorgangId) {
		return OzgCloudUploadFile.builder()
		  .fileName(downloadAnhangDetails.getDatei())
		  .fieldName(POSTFACH)
		  .vorgangId(vorgangId)
		  .contentType(downloadAnhangDetails.getMimeType()).build();
	}

	private static String getVorgangId(final DownloadAnhangDetails downloadAnhangDetails) {
		try {
			if (Objects.nonNull(downloadAnhangDetails) && StringUtils.isNotEmpty(downloadAnhangDetails.getDateiBezeichnung())) {
				var dateiBezeichnung = downloadAnhangDetails.getDateiBezeichnung();
				return dateiBezeichnung.substring(VORGANG_ID_REFIX.length(), VORGANG_ID_REFIX.length() + ObjectId.get().toHexString().length());
			}
		} catch (Exception e) {
			LOG.error("Error extracting VorgangId from Dateibezeichnung. Returning null");
		}

		return null;
	}

	public AuftragAngelegt send(PostfachNachricht nachricht) {
		var auftrag = createBereitstellungsAuftrag(nachricht);

		return restClient.postBereitstellungsauftrag(auftrag);
	}

	BereitstellungAuftragNeuV3 createBereitstellungsAuftrag(PostfachNachricht nachricht) {
		var bereitstellungAuftrag = mapper.toTransferMessage(nachricht);
		bereitstellungAuftrag.setAbsender(getAbsender());

		if (anhangNeedToBeUploaded(bereitstellungAuftrag)) {
			bereitstellungAuftrag.setAnhaenge(
			  nachricht.getAttachments().stream().map(fileId -> uploadAnhang(fileId, nachricht.getVorgangId())).toList());
		}

		if (isMitteilung(bereitstellungAuftrag)) {
			bereitstellungAuftrag.setBescheidDatum(null);
		}

		if (isReplyPossible(nachricht)) {
			bereitstellungAuftrag.setAntwort(createAntwort(nachricht));
		}

		return bereitstellungAuftrag;
	}

	String getAbsender() {
		return postfachProperties.getSender();
	}

	BereitstellungAnhangWithUploadId uploadAnhang(String fileId, String vorgangId) {
		var file = attachmentService.getCloudFile(fileId);
		try {
			var uploadInfo = uploadFile(file);
			var bereitstellungAnhang = fileMapper.toBereitstellungAnhangWithUploadId(file);
			bereitstellungAnhang.setDateiBezeichnung(vorgangId);
			bereitstellungAnhang.setUploadId(uploadInfo.getUploadId());

			return bereitstellungAnhang;
		} catch (IOException e) {
			var message = String.format("Error uploading attachment id [%s] to ELSTER Transfer", fileId);
			LOG.error(message, e);
			throw new TechnicalException(message, e);
		}
	}

	UploadInfo uploadFile(OzgCloudFile file) throws IOException {
		Path tempFile = TempFileUtils.createTmpFile();
		try (FileOutputStream fileOutputStream = new FileOutputStream(tempFile.toFile())) {
			attachmentService.getFileContent(file.getId(), fileOutputStream);
			return restClient.postUpload(tempFile);
		} finally {
			Files.deleteIfExists(tempFile);
		}
	}

	BereitstellungAntwort createAntwort(PostfachNachricht nachricht) {
		var antwort = new BereitstellungAntwort();
		// FIXME: nachricht needs to have a nachricht id already set. Needs to be fixed in de.ozgcloud.nachrichten.postfach.PostfachService or de.ozgcloud.nachrichten.postfach.PostfachEventListener.buildNachricht
		if (Objects.isNull(nachricht.getId())) {
			antwort.setZuordnungskriterium(Objects.isNull(nachricht.getId()) ? UUID.randomUUID() : ObjectIdToUUIDConverter.toUUID(nachricht.getId()));
			antwort.setAntwortBetreffe(
			  Set.of(VORGANG_ID_REFIX + nachricht.getVorgangId() + "]" + postfachProperties.getReplySubjectPrefix() + nachricht.getSubject()));
		} else {
			antwort.setZuordnungskriterium(ObjectIdToUUIDConverter.toUUID(nachricht.getId()));
			antwort.setAntwortBetreffe(Set.of(postfachProperties.getReplySubjectPrefix() + nachricht.getSubject()));
		}

		antwort.setAntwortFrist(LocalDate.now().plusDays(postfachProperties.getReplyInDays()));

		return antwort;
	}

	public BereitstellungInfo getStatusOfBereitstellungsauftrag(String id) {
		var url = String.format("%s%s/%s", postfachProperties.getServer(), BEREITSTELLUNGS_ENDPUNKT, id);
		return restClient.getForObject(url, BereitstellungInfo.class);
	}

	public Stream<PostfachNachricht> getAllMessages() {
		var antworten = loadAntworten();

		/*return antworten.getDatensaetze().stream()
		  .map(this::getAntwort).filter(nachricht -> nachricht.getVorgangId() != null);*/
		LOG.warn("Saving of Antworten is disabled right now, returning an empty stream");
		return Stream.of();
	}

	public AntwortDetailsList loadAntworten() {
		return restClient.getForObject(postfachProperties.getServer() + ANTWORTEN_ENDPOINT, AntwortDetailsList.class);
	}

	PostfachNachricht getAntwort(AntwortDetails antwortDetails) {
		LOG.info("Processing AntwortDetail {} with DownloadDate {}", antwortDetails, antwortDetails.getDownloadZeitpunkt());

		var nachrichtBuilder = PostfachNachricht.builder()
		  .direction(PostfachNachricht.Direction.IN)
		  .subject(postfachProperties.getReplySubjectPrefix())
		  .mailBody(postfachProperties.getBody());

		if (Objects.nonNull(antwortDetails.getZuordnungskriterium()) && CollectionUtils.isNotEmpty(antwortDetails.getAnhaenge())) {
			nachrichtBuilder
			  .replyOption(PostfachNachricht.ReplyOption.POSSIBLE);

			// TODO update to new API as soon the new api is available.
			var vorgangId = getVorgangId(antwortDetails.getAnhaenge().getFirst());
			if (Objects.nonNull(vorgangId)) {
				nachrichtBuilder.vorgangId(vorgangId).attachments(getAttachmentIds(antwortDetails, new ObjectId(vorgangId)));
			} else {
				var nachrichtId = ObjectIdToUUIDConverter.toObjectId(antwortDetails.getZuordnungskriterium());
				LOG.warn("This nachricht with AntwortDetails {} will be ignored right now. FIXME when new api is available", antwortDetails);
				nachrichtBuilder.referencedNachricht(nachrichtId.toHexString()).attachments(getAttachmentIds(antwortDetails, nachrichtId));
			}
		}

		return nachrichtBuilder.build();
	}

	// TODO update to new API as soon the new api is available.
	private List<String> getAttachmentIds(final AntwortDetails antwortDetails, final ObjectId nachrichtId) {
		List<String> attachmentIds = new ArrayList<>();
		if (Objects.nonNull(antwortDetails.getAnhaenge())) {
			var idList = antwortDetails.getAnhaenge().stream()
			  .filter(Objects::nonNull)
			  .map(downloadAnhangDetails -> saveAnhang(downloadAnhang(downloadAnhangDetails), downloadAnhangDetails, nachrichtId))
			  .filter(Optional::isPresent)
			  .map(Optional::get)
			  .toList();

			attachmentIds.addAll(idList);
		}

		return attachmentIds;
	}

	byte[] downloadAnhang(DownloadAnhangDetails downloadAnhangDetails) {
		var url = postfachProperties.getServer() + downloadAnhangDetails.getDownloadUrl();

		return restClient.getForObject(url, byte[].class);
	}

	Optional<String> saveAnhang(byte[] data, DownloadAnhangDetails downloadAnhangDetails, ObjectId nachrichtId) {
		LOG.info("Saving Attachment from DownloadAnhangDetails {}", downloadAnhangDetails);
		Optional<String> fileIdOptional = Optional.empty();
		if (Objects.nonNull(data)) {
			// TODO update to new API as soon the new api is available. Save to file here and then upload attachment
			try (var dataIn = new ByteArrayInputStream(data)) {
				if (Objects.isNull(nachrichtId)) {
					// can be removed when new api is available
					var vorgangId = getVorgangId(downloadAnhangDetails);
					if (Objects.nonNull(vorgangId)) {
						fileIdOptional = storeToOzgCloud(downloadAnhangDetails, vorgangId, dataIn);
					}
				} else {
					// use new api here
					// fileIdOptional = storeToOzgCloud(downloadAnhangDetails, nachrichtId, dataIn);
					LOG.warn("Saving with NachrichtId not implemented yet.");
				}
			} catch (IOException e) {
				LOG.error("Error transferring attachment body from Elster-Transfer to OZG-Cloud.", e);
				throw new TechnicalException("Error downloading anhang", e);
			}
		} else {
			LOG.warn("DownloadAnhangDetails {} available, but there is no content at {}", downloadAnhangDetails,
			  downloadAnhangDetails.getDownloadUrl());
		}

		return fileIdOptional;
	}

	Optional<String> storeToOzgCloud(DownloadAnhangDetails downloadAnhangDetails, String vorgangId, ByteArrayInputStream dataIn) throws IOException {
		/*OzgCloudFileId ozgCloudFileId;

		try (BufferedInputStream dataStream = new BufferedInputStream(dataIn)) {
			ozgCloudFileId = attachmentService.store(createOzgCloudUploadFile(downloadAnhangDetails, vorgangId), dataStream);
		}

		if (Objects.nonNull(ozgCloudFileId)) {
			return Optional.of(ozgCloudFileId.toString());
		}

		LOG.error("Received no OzgCloudFileId when saving attachment {} to vorgangId {}", downloadAnhangDetails, vorgangId);
		 */
		LOG.warn("Saving attachment {} to vorgangId {} is disabled right now", downloadAnhangDetails, vorgangId);
		return Optional.empty();
	}
}


